/* M. Beeson 5.16.04  
Typechecking for Otter-lambda
Typechecking is turned on by set(types), and there should be 
a list(types) in the input file containing terms of the form 
type(returntype,  f(arg0type, arg1type,...)).
*/   

#include <assert.h>
#include "header.h"
#include "unify.h"    /* BIND and DEREFERENCE */
#include "bterms.h"
#include "bsym.h"
#include "beta.h"
/*_______________________________________________________________*/
static int possible_type(struct term *rType, struct term *t, 
                         int parent,int argNumber,struct context *c)
/* return 1 if according to the Types list, t can occur with 
value type rType as the argNumber-th argument of a complex term with sym_num parent. 
Presumes Types is not NULL.
*/
{  struct clause *p;
   struct literal *l;
   struct term *s;
   struct term *returnType;
   struct term *value;
   static int forgotten[5];
   static int nforgotten;
   int k;
   struct rel *r,*r2; 
#if 0   
           fprintf(stdout, "can "); // DEBUG
           print_term_nl(stdout,t); // DEBUG
           fprintf(stdout, "have type "); // DEBUG
           print_term_nl(stdout,rType);  // DEBUG
#endif             
   if(t->type == COMPLEX)
     { for(p = Types->first_cl; p; p = p->next_cl)
         { l = p->first_lit;
           s = l->atom;
           returnType = ARG0(s);   
           value = ARG1(s);
          
           if(returnType != rType || value->type != COMPLEX || value->sym_num != t->sym_num )
              continue;
           r = t->farg;
           r2 = value->farg;
           for(k=0;r && r2;k++,r=r->narg,r2=r2->narg)
              { if(!possible_type(r2->argval,r->argval,t->sym_num,k,c))
                    break;
              }
           if(r || r2)
              continue;
           return 1;
        }
      return 0;
    }
  if(t->type == NAME)
    { int found = 0;
      for(p = Types->first_cl; p; p = p->next_cl)
         { l = p->first_lit;
           s = l->atom;
           returnType = ARG0(s);   
           value = ARG1(s);
           if(value->type != NAME)
              continue;
           if( t->sym_num == value->sym_num)
              { if(returnType == rType)
                   return 1;
                found = 1;
              }
         }
      if(!found)
         { for(k=0;k<nforgotten;k++)
              { if(forgotten[k] == t->sym_num)
                  return 0;
              }
           fprintf(stderr, "\nYou have set(types), but you did not assign a type to the constant ");
           print_term_nl(stderr,t);
           if(nforgotten < 5)
              { forgotten[k] = t->sym_num;
                ++nforgotten;
              }
         }
      return 0;
    }     
  if(t->type == VARIABLE && c->terms[t->varnum])
    return possible_type(rType,c->terms[t->varnum],parent,argNumber,c->contexts[t->varnum]);
  if(t->type == VARIABLE )
    return 1;  // for now--FIX THIS
  return 0;   
}     
/*_______________________________________________________________*/
int type_check(struct term *t, int parent, int argNumber, struct context *c)
/* called when set(types) command is in effect, so Types should be 
non-NULL (but isn't required to be).
  Return 1 if t can occur as an argNumber-th argument of sym_num parent,
interpreting variables that occur in t as belong to context c.
*/
/*  This could be made more efficient by use of a hash table.
*/

{  struct clause *p;
   struct literal *l;
   struct term *s;
   struct term *returnType;
   struct term *value;
  // fprintf(stdout, "Type checking: ");  // DEBUG
  // print_term_nl(stdout,t);                    // DEBUG
   if(!Types)
      { abend("You have set(types) and set(lambda) but no list(types).  Lambda-unification will therefore always fail.");
        return 0;
      }
   for(p = Types->first_cl; p; p = p->next_cl)
      { l = p->first_lit;
        s = l->atom;
        if(s->type != COMPLEX || s->farg == NULL || s->farg->narg == NULL)
           abend("Invalid entry in types list; entries must match type(a,b).");
        value = ARG1(s);
        if(value->type != COMPLEX)
           continue;  // a term like type(int,c)  will be used in possible_type, not here
        returnType = ARG0(value);
        if(t->type  != value->type)
           continue;
        if(value->type != COMPLEX)
           continue;
        if(parent != value->sym_num)
           continue;
        if(possible_type(returnType,t,parent,argNumber,c))
           return 1;
      }  
   // fprintf(stdout,"Type checking failed on.\n");  // DEBUG      
   return 0;
}           
        
/*____________________________________________________*/ 
 int getConstantType(struct term *t)
 // assumes t is a constant; return the syn_num of its type
 // or return -1 if t has no type in the Types list.
 { struct clause *c; 
   int f;
   char *p;
   struct term *s;
   if(t->type != NAME)
      assert(0);
   f = str_to_sn("type",2);
   if(Types == NULL)
      { if(Flags[TYPES_FLAG].val)
           assert(0);  // not supposed to call this unless Flags[TYPEFLAG].val     
        else
           abend("You have set(types) but no list(types).");
      }
   for(c = Types->first_cl; c ; c = c->next_cl)
      { s = c->first_lit->atom;
        if(s->sym_num != f || ARG1(s)->sym_num != t->sym_num)
           continue;
        return ARG0(s)->sym_num;
      }
   p = sn_to_str(t->sym_num);
   fprintf(stderr,"You haven't recorded the type of constant %s in the types list. You may be missing possible inferences.\n",p);
   return -1;  // t has no recorded type   
 }        
